import numpy as np
from matplotlib import pyplot as plt

from decision_tree.node import DecisionTreeNode


def plot_roc_curve(y_test, p_pred, save_path=None):
    positive_samples = sum(1 for y in y_test if y == 0)
    tpr = []
    fpr = []
    for w in np.arange(-0.01, 1.02, 0.01):
        y_pred = [(0 if p.get(0, 0) > w else 1) for p in p_pred]
        tpr.append(sum(1 for yp, yt in zip(y_pred, y_test) if yp == 0 and yt == 0) / positive_samples)
        fpr.append(sum(1 for yp, yt in zip(y_pred, y_test) if yp == 0 and yt != 0) / (len(y_test) - positive_samples))
    plt.figure(figsize=(7, 7))
    plt.plot(fpr, tpr)
    plt.plot([0, 1], [0, 1], linestyle="--")
    plt.xlabel("False positive rate")
    plt.ylabel("True positive rate")
    plt.xlim(-0.01, 1.01)
    plt.ylim(-0.01, 1.01)
    plt.tight_layout()
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()


def rectangle_bounds(bounds):
    return ((bounds[0][0], bounds[0][0], bounds[0][1], bounds[0][1]),
            (bounds[1][0], bounds[1][1], bounds[1][1], bounds[1][0]))


def plot_2d_tree(tree_root, bounds, colors):
    if isinstance(tree_root, DecisionTreeNode):
        if tree_root.split_dim:
            plot_2d_tree(tree_root.left, [bounds[0], [bounds[1][0], tree_root.split_value]], colors)
            plot_2d_tree(tree_root.right, [bounds[0], [tree_root.split_value, bounds[1][1]]], colors)
            plt.plot(bounds[0], (tree_root.split_value, tree_root.split_value), c=(0, 0, 0))
        else:
            plot_2d_tree(tree_root.left, [[bounds[0][0], tree_root.split_value], bounds[1]], colors)
            plot_2d_tree(tree_root.right, [[tree_root.split_value, bounds[0][1]], bounds[1]], colors)
            plt.plot((tree_root.split_value, tree_root.split_value), bounds[1], c=(0, 0, 0))
    else:
        x, y = rectangle_bounds(bounds)
        plt.fill(x, y, c=colors[tree_root.y] + [0.2])


def plot_2d(tree, X, y, save_path=None):
    plt.figure(figsize=(9, 9))
    colors = dict((c, list(np.random.random(3))) for c in np.unique(y))
    bounds = list(zip(np.min(X, axis=0), np.max(X, axis=0)))
    plt.xlim(*bounds[0])
    plt.ylim(*bounds[1])
    plot_2d_tree(tree.root, list(zip(np.min(X, axis=0), np.max(X, axis=0))), colors)
    for c in np.unique(y):
        plt.scatter(X[y == c, 0], X[y == c, 1], c=[colors[c]], label=c)
    plt.legend()
    plt.tight_layout()
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()



import os.path

from sklearn.datasets import make_moons

from decision_tree.classifier import DecisionTreeClassifier
from decision_tree.draw_tree import draw_tree
from decision_tree.plots import plot_2d, plot_roc_curve


def main():
    noise = 0.35
    X, y = make_moons(1500, noise=noise)
    X_test, y_test = make_moons(200, noise=noise)
    tree = DecisionTreeClassifier(max_depth=5, min_samples_leaf=30)
    tree.fit(X, y)
    if not os.path.exists('../plots'):
        os.mkdir('../plots')
    plot_2d(tree, X, y, save_path='../plots/2d.png')
    plot_roc_curve(y_test, tree.predict_proba(X_test), save_path='../plots/roc_curve.png')
    draw_tree(tree, save_path='../plots/tree.png')


if __name__ == "__main__":
    main()



from collections import defaultdict
from typing import Callable

import numpy as np


def gini(x: np.ndarray) -> float:
    """
    Считает коэффициент Джини для массива меток x.
    """
    m = defaultdict(int)
    for y in x:
        m[y] += 1
    m = np.array(list(m.values())) / x.shape[0]
    res = m.dot(1 - m)
    # print("Giny=", res)
    return res


def entropy(x: np.ndarray) -> float:
    """
    Считает энтропию для массива меток x.
    """
    m = defaultdict(int)
    for y in x:
        m[y] += 1
    m = np.array(list(m.values())) / x.shape[0]
    res = -np.sum(m * np.log2(m))
    # print("Entropy=", res)
    return res


def gain(left_y: np.ndarray, right_y: np.ndarray, criterion: Callable) -> float:
    """
    Считает информативность разбиения массива меток.

    Parameters
    ----------
    left_y : np.ndarray
        Левая часть разбиения.
    right_y : np.ndarray
        Правая часть разбиения.
    criterion : Callable
        Критерий разбиения.
    """
    res = criterion(np.concatenate([left_y, right_y])) * (len(left_y) + len(right_y)) - len(right_y) * criterion(
        right_y) - len(left_y) * criterion(left_y)
    return res


from typing import Optional, NoReturn, List, Any, Dict, Union

import numpy as np

from decision_tree.metrics import gini, entropy, gain
from decision_tree.node import DecisionTreeLeaf, DecisionTreeNode


class DecisionTreeClassifier:
    """
    Attributes
    ----------
    root : Union[DecisionTreeNode, DecisionTreeLeaf]
        Корень дерева.

    (можете добавлять в класс другие аттрибуты).

    """

    def __init__(self, criterion: str = "gini",
                 max_depth: Optional[int] = None,
                 min_samples_leaf: int = 1):
        """
        Parameters
        ----------
        criterion : str
            Задает критерий, который будет использоваться при построении дерева.
            Возможные значения: "gini", "entropy".
        max_depth : Optional[int]
            Ограничение глубины дерева. Если None - глубина не ограничена.
        min_samples_leaf : int
            Минимальное количество элементов в каждом листе дерева.

        """
        self.root = None
        self.criterion = gini if criterion == "gini" else entropy
        self.max_depth = max_depth
        self.min_samples_leaf = min_samples_leaf

    def fit(self, X: np.ndarray, y: np.ndarray) -> NoReturn:
        """
        Строит дерево решений по обучающей выборке.

        Parameters
        ----------
        X : np.ndarray
            Обучающая выборка.
        y : np.ndarray
            Вектор меток классов.
        """
        self.root = self._build(X, y, 0)

    def predict_proba(self, X: np.ndarray) -> List[Dict[Any, float]]:
        """
        Предсказывает вероятность классов для элементов из X.

        Parameters
        ----------
        X : np.ndarray
            Элементы для предсказания.

        Return
        ------
        List[Dict[Any, float]]
            Для каждого элемента из X возвращает словарь
            {метка класса -> вероятность класса}.
        """
        ans = []
        for x in X:
            cur = self.root
            while not isinstance(cur, DecisionTreeLeaf):
                if x[cur.split_dim] < cur.split_value:
                    cur = cur.left
                else:
                    cur = cur.right
            ans.append(cur.m)
        return ans

    def predict(self, X: np.ndarray) -> list:
        """
        Предсказывает классы для элементов X.

        Parameters
        ----------
        X : np.ndarray
            Элементы для предсказания.

        Return
        ------
        list
            Вектор предсказанных меток для элементов X.
        """
        proba = self.predict_proba(X)
        return [max(p.keys(), key=lambda k: p[k]) for p in proba]

    def _build(self, X: np.ndarray, y: np.ndarray, depth: int) -> Union['DecisionTreeNode', DecisionTreeLeaf]:
        if len(np.unique(y)) == 1:
            return DecisionTreeLeaf(y)

        if depth == self.max_depth:
            return DecisionTreeLeaf(y)

        if len(y) < 2 * self.min_samples_leaf:
            return DecisionTreeLeaf(y)

        dim = -1
        threshold = -1
        inf_gain = 0

        for i in range(X.shape[1]):
            ind = list(range(X.shape[0]))
            ind.sort(key=lambda k: X[k][i])
            for j in range(self.min_samples_leaf, X.shape[0] - self.min_samples_leaf):
                cur_gain = gain(y[ind[:j]], y[ind[j:]], self.criterion)
                if cur_gain > inf_gain:
                    dim = i
                    threshold = X[ind[j]][i]
                    inf_gain = cur_gain

        if dim == -1:
            return DecisionTreeLeaf(y)

        left = [i for i, (x, _) in enumerate(zip(X, y)) if x[dim] < threshold]
        right = [i for i, (x, _) in enumerate(zip(X, y)) if x[dim] >= threshold]

        return DecisionTreeNode(dim, threshold,
                                self._build(X[left], y[left], depth + 1),
                                self._build(X[right], y[right], depth + 1)
                                )



from collections import defaultdict
from typing import Union


class DecisionTreeLeaf:
    """

    Attributes
    ----------
    y : Тип метки (напр., int или str)
        Метка класса, который встречается чаще всего среди элементов листа дерева
    """

    def __init__(self, x):
        m = defaultdict(int)
        for y in x:
            m[y] += 1
        self.m = {k: v / len(x) for k, v in m.items()}
        self.y = sorted(list(m.keys()), key=lambda x: -m[x])[0]


class DecisionTreeNode:
    """

    Attributes
    ----------
    split_dim : int
        Измерение, по которому разбиваем выборку.
    split_value : float
        Значение, по которому разбираем выборку.
    left : Union[DecisionTreeNode, DecisionTreeLeaf]
        Поддерево, отвечающее за случай x[split_dim] < split_value.
    right : Union[DecisionTreeNode, DecisionTreeLeaf]
        Поддерево, отвечающее за случай x[split_dim] >= split_value.
    """

    def __init__(self, split_dim: int, split_value: float,
                 left: Union['DecisionTreeNode', DecisionTreeLeaf],
                 right: Union['DecisionTreeNode', DecisionTreeLeaf]):
        self.split_dim = split_dim
        self.split_value = split_value
        self.left = left
        self.right = right



from decision_tree.node import DecisionTreeNode
import matplotlib.pyplot as plt


def tree_depth(tree_root):
    if isinstance(tree_root, DecisionTreeNode):
        return max(tree_depth(tree_root.left), tree_depth(tree_root.right)) + 1
    else:
        return 1


def draw_tree_rec(tree_root, x_left, x_right, y):
    x_center = (x_right - x_left) / 2 + x_left
    if isinstance(tree_root, DecisionTreeNode):
        x_center = (x_right - x_left) / 2 + x_left
        x = draw_tree_rec(tree_root.left, x_left, x_center, y - 1)
        plt.plot((x_center, x), (y - 0.1, y - 0.9), c=(0, 0, 0))
        x = draw_tree_rec(tree_root.right, x_center, x_right, y - 1)
        plt.plot((x_center, x), (y - 0.1, y - 0.9), c=(0, 0, 0))
        plt.text(x_center, y, "x[%i] < %f" % (tree_root.split_dim, tree_root.split_value),
                 horizontalalignment='center')
    else:
        plt.text(x_center, y, str(tree_root.y),
                 horizontalalignment='center')
    return x_center


def draw_tree(tree, save_path=None):
    td = tree_depth(tree.root)
    plt.figure(figsize=(0.33 * 2 ** td, 2 * td))
    plt.xlim(-1, 1)
    plt.ylim(0.95, td + 0.05)
    plt.axis('off')
    draw_tree_rec(tree.root, -1, 1, td)
    plt.tight_layout()
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()
